use context::prelude::*;

use std::{
    sync::{
        Arc,
        mpsc::{Receiver, channel},
    },
    thread,
};

use anyhow::Result;

pub struct LoadingScreen<T: Send + Sync + 'static> {
    result: Option<T>,
    receiver: Receiver<T>,
}

impl<T: Send + Sync + 'static> LoadingScreen<T> {
    pub fn load<R, L, G>(gui: Arc<G>, loader: L) -> Result<Self>
    where
        R: Fn(T) -> Result<()> + Send + Sync + 'static,
        T: Send + Sync + 'static,
        L: FnOnce() -> T + Send + Sync + 'static,
        G: TopLevelGui + TopGui + Send + Sync + 'static,
    {
        gui.enable()?;

        let (sender, receiver) = channel();

        thread::spawn(move || {
            let _ = sender.send(loader());
            let _ = gui.disable();
        });

        Ok(Self {
            result: None,
            receiver,
        })
    }

    pub fn check_ready(&mut self) -> bool {
        if self.result.is_some() {
            return true;
        }

        match self.receiver.try_recv() {
            Ok(result) => {
                self.result = Some(result);
                true
            }
            Err(_) => false,
        }
    }

    /// Will panic if there is no result.
    /// Use `check_ready` before calling this.
    pub fn take_result(mut self) -> T {
        self.result.take().unwrap()
    }
}
